--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      Code Editor - main class
-- @author    Valentin Schmidt
-- @version   0.5
--****************************************************************************

global $

property pAppName

-- WIDGETS
property pMainWindow
property pEditor
property pMenuBar

-- ACTIONS
property pActionSave
property pActionCut
property pActionCopy
property pActionPaste

-- OTHER QT OBJECTS
property pClipboard
property pPrinter

-- OTHER VARS
property pFileName
property pIconPath
property pLangDict
property pDefaultFont

----------------------------------------
-- @constructor
----------------------------------------
on new (me)

	me.pAppName = "CodeEditor"

	-- INIT VARS
	me.pLangDict = [:]
	me.pLangDict["c"] = "CPP"
	me.pLangDict["cpp"] = "CPP"
	me.pLangDict["h"] = "CPP"
	me.pLangDict["hpp"] = "CPP"
	me.pLangDict["js"] = "JavaScript"
	me.pLangDict["lua"] = "Lua"
	me.pLangDict["ls"] = "Lua"
	me.pLangDict["lsw"] = "Lua"
	me.pLangDict["py"] = "Python"
	me.pLangDict["pyw"] = "Python"

	me.pIconPath = $.@("images/"&$.os&"/")

	me.pDefaultFont = $.Qt.Font.new()
	me.pDefaultFont.setFamily("Courier New")
	me.pDefaultFont.setPointSize(10.0)

	-- CREATE MAIN WINDOW
	me.pMainWindow = $.Qt.MainWindow.new()
	me.pMainWindow.connect($.Qt.Event.Close, me, #fileQuit)
	me.pMainWindow.setWindowTitle(me.pAppName)
	me.pMainWindow.setWindowIcon($.@("images/logo.png"))
	me.pMainWindow.setToolButtonStyle($.Qt.ToolButtonFollowStyle)

	-- ADD DROP SUPPORT
	$.Qt.registerDropHandler(me.pMainWindow, #dropHandler, me)
	me.pMainWindow.setAcceptDrops(TRUE)

	-- CREATE QSCINTILLA
	me.pEditor = $.Qt.Scintilla.new(me.pMainWindow)
	me.pMainWindow.setCentralWidget(me.pEditor)

	me.pEditor.setFont(me.pDefaultFont)
	me.pEditor.setMarginsFont(me.pDefaultFont)
	me.pEditor.setMarginWidth(0, "00000")
	me.pEditor.setFolding($.Qt.QsciScintilla.BoxedTreeFoldStyle)

	me.pEditor.setMarginLineNumbers(0, true)
	me.pEditor.setBraceMatching($.Qt.QsciScintilla.SloppyBraceMatch)
	me.pEditor.setAutoIndent(true)
	me.pEditor.setCaretLineVisible(true)
	me.pEditor.setCaretLineBackgroundColor(rgb("C4E8FD"))
	me.pEditor.setTabWidth(2)
	me.pEditor.setTabIndents(true)

	-- CREATE MENUBAR
	me.pMenuBar = $.Qt.MenuBar.new(me.pMainWindow)
	me.pMainWindow.setMenuBar(me.pMenuBar)

	me.setupFileActions()
	me.setupEditActions()
	me.setupHelpActions()

	-- CREATE STATUSBAR
	statusBar = $.Qt.StatusBar.new(pMainWindow)
	me.pMainWindow.setStatusBar(statusBar)

	-- WATCH SYSTEM CLIPBOARD FOR CHANGES
	--me.pClipboard = $.Qt.Clipboard.new(me.pMainWindow)
	me.pClipboard = $.Qt.Application.getClipboard()
	me.pClipboard.connect(#dataChanged, me, #clipboardDataChanged)

	-- CREATE A PRINTER OBJECT
	me.pPrinter = $.Qt.SciPrinter.new(0, $.Qt.QPrinter.HighResolution)

	-- SETUP TEXTEDIT SIGNALS
	-- me.pEditor.connect(#cursorPositionChanged, me, #cursorPositionChanged)
	me.pEditor.connect(#copyAvailable, me, #copyAvailable)
	me.pEditor.connect(#modificationChanged, me, #modificationChanged) -- ???

	-- LOAD FILE?
	if $.args.count>1 then me.load($.args[2])
	else me.fileNew()

	-- SET CWD
	cwd = $.path -- use LSW script dir
	$.Qt.Dir.new(#Dir).setCurrent(cwd)

	-- RESIZE AND SHOW MAIN WINDOW
	settings = $.Qt.Settings.new("Fluxus", me.pAppName, $.Qt.QSettings.IniFormat)
	state = value(settings.value("mainwindow/state"))
	if state=2 then
		me.pMainWindow.resize(680, 680)
		me.pMainWindow.setWindowState(state)
	else
		rect = value(settings.value("mainwindow/rect"))
		if ilk(rect)=#rect then me.pMainWindow.setGeometry(rect)
		else me.pMainWindow.resize(680, 680)
	end if
	me.pMainWindow.show()

	return me
end

--------------------------------------
--
--------------------------------------
on setupFileActions (me)

	toolBar = $.Qt.ToolBar.new(me.pMainWindow)
	toolBar.setWindowTitle("File Actions")
	toolBar.setIconSize(16, 16)
	me.pMainWindow.addToolBar(toolBar)

	menu = me.pMenuBar.addMenu("&File")

	action = $.Qt.Action.new(menu)
	action.setText("&New")
	action.setIcon(me.pIconPath & "filenew.png")
	action.connect(#triggered, me, #fileNew)
	action.setShortcut("Ctrl+N")
	toolBar.addAction(action)
	menu.addAction(action)

	action = $.Qt.Action.new(menu)
	action.setText("&Open")
	action.setIcon(me.pIconPath & "fileopen.png")
	action.connect(#triggered, me, #fileOpen)
	action.setShortcut("Ctrl+O")
	toolBar.addAction(action)
	menu.addAction(action)

	menu.addSeparator()

	action = $.Qt.Action.new(menu)
	action.setText("&Save")
	action.setIcon(me.pIconPath & "filesave.png")
	action.connect(#triggered, me, #fileSave)
	action.setShortcut("Ctrl+S")
	toolBar.addAction(action)
	menu.addAction(action)
	me.pActionSave = action

	action = $.Qt.Action.new(menu)
	action.setText("Save &As...")
	action.setIcon(me.pIconPath & "filesave.png")
	action.connect(#triggered, me, #fileSaveAs)
	action.setShortcut("Ctrl+Shift+S")
	menu.addAction(action)

	menu.addSeparator()

	action = $.Qt.Action.new(menu)
	action.setText("E&xport PDF...")
	action.setIcon(me.pIconPath & "exportpdf.png")
	action.connect(#triggered, me, #filePrintPdf)
	toolBar.addAction(action)
	menu.addAction(action)

	menu.addSeparator()

	action = $.Qt.Action.new(menu)
	action.setText("Page Setup")
	action.connect(#triggered, me, #filePageSetup)
	menu.addAction(action)

	action = $.Qt.Action.new(menu)
	action.setText("Print Preview")
	action.connect(#triggered, me, #filePrintPreview)
	menu.addAction(action)

	action = $.Qt.Action.new(menu)
	action.setText("&Print...")
	action.setIcon(me.pIconPath & "fileprint.png")
	action.connect(#triggered, me, #filePrint)
	action.setShortcut("Ctrl+P")
	toolBar.addAction(action)
	menu.addAction(action)

	menu.addSeparator()

	action = $.Qt.Action.new(menu)
	action.setText("&Quit")
	action.connect(#triggered, me, #fileQuit)
	menu.addAction(action)
end

--------------------------------------
--
--------------------------------------
on setupEditActions (me)

	toolBar = $.Qt.ToolBar.new(me.pMainWindow)
	toolBar.setWindowTitle("Edit Actions")
	toolBar.setIconSize(16, 16)
	me.pMainWindow.addToolBar(toolBar)
	toolBar.setVisible(FALSE)

	menu = pMenuBar.addMenu("&Edit")

	action = $.Qt.Action.new(menu)
	action.setText("&Undo")
	action.setIcon(me.pIconPath & "editundo.png")
	action.connect(#triggered, me.pEditor, #undo)
	action.setShortcut("Ctrl+Z")
	toolBar.addAction(action)
	menu.addAction(action)

	action = $.Qt.Action.new(menu)
	action.setText("&Redo")
	action.setIcon(me.pIconPath & "editredo.png")
	action.connect(#triggered, me.pEditor, #redo)
	action.setShortcut("Ctrl+Y")
	toolBar.addAction(action)
	menu.addAction(action)

	menu.addSeparator()

	action = $.Qt.Action.new(menu)
	action.setText("Cu&t")
	action.setIcon(me.pIconPath & "editcut.png")
	action.connect(#triggered, me.pEditor, #cut)
	action.setShortcut("Ctrl+X")
	toolBar.addAction(action)
	menu.addAction(action)
	action.setEnabled(FALSE)
	me.pActionCut = action

	action = $.Qt.Action.new(menu)
	action.setText("&Copy")
	action.setIcon(me.pIconPath & "editcopy.png")
	action.connect(#triggered, me.pEditor, #copy)
	action.setShortcut("Ctrl+C")
	toolBar.addAction(action)
	menu.addAction(action)
	action.setEnabled(FALSE)
	me.pActionCopy = action

	action = $.Qt.Action.new(menu)
	action.setText("&Paste")
	action.setIcon(me.pIconPath & "editpaste.png")
	action.connect(#triggered, me.pEditor, #paste)
	action.setShortcut("Ctrl+V")
	toolBar.addAction(action)
	menu.addAction(action)
	me.pActionPaste = action

end

--------------------------------------
--
--------------------------------------
on setupHelpActions (me)

	menuHelp = me.pMenuBar.addMenu("Help")

	action = $.Qt.Action.new(menuHelp)
	action.setText("Edit me!")
	action.connect(#triggered, me, #helpEdit)
	menuHelp.addAction(action)

	menuHelp.addSeparator()

	action = $.Qt.Action.new(menuHelp)
	action.setText("About &"&me.pAppName)
	action.connect(#triggered, me, #helpAbout)
	menuHelp.addAction(action)

	action = $.Qt.Action.new(menuHelp)
	action.setText("About &Qt")
	action.connect(#triggered, me, #helpAboutQt)
	menuHelp.addAction(action)
 end

--------------------------------------
--
--------------------------------------
on load (me, fn)
	data = $.file.getBytes(fn)
	if data=false then return FALSE

	-- GUESS ENCODING
	enc = $.bytes.hasBOM(data)
	if enc=false then
		if $.bytes.isUtf8(data) then enc = "utf-8"
	end if
	if enc=false then enc = "windows-1252"
	data.position = 1

	-- LOAD A LEXER?
	od = _player.itemdelimiter
	_player.itemdelimiter = "."
	ftype = $.string.toLower(the last item of fn)
	_player.itemdelimiter = od
	lang = me.pLangDict[ftype]
	--put lang
	
	if not voidP(lang) then
		
		lexerName = symbol("Lexer"&lang)
		lexer = $.Qt.SciLexer.new(me.pEditor, lexerName)
		me.pEditor.setLexer(lexer)

		if lang = #Lua then
			me.pEditor.setFolding(0) -- tmp hack: deactivate wrong lua folding for lingo
		else
			me.pEditor.setFolding($.Qt.QsciScintilla.BoxedTreeFoldStyle)
		end if
	end if

	me.pEditor.setText(data.readRawString(data.length, enc))

	me.setCurrentFileName(fn)
	return TRUE
end

--------------------------------------
--
--------------------------------------
on maybeSave (me)
	if not me.pEditor.isModified() then return TRUE

	tStandardButtons = bitOr(bitOr($.Qt.QMessageBox.Save, $.Qt.QMessageBox.Discard), $.Qt.QMessageBox.Cancel)
	msg = "The document has been modified."&$.LF&"Do you want to save your changes?"
	ret = $.Qt.MessageBox.warning(me.pMainWindow, me.pAppName,msg,tStandardButtons)

	if ret = $.Qt.QMessageBox.Save then return me.fileSave()
	if ret = $.Qt.QMessageBox.Cancel then return FALSE

	return TRUE
end

--------------------------------------
--
--------------------------------------
on setCurrentFileName (me, fn)
	me.pFileName = fn
	me.pEditor.setModified(FALSE)
	if voidP(fn) then
		shownName = "untitled.txt"
	else
		-- FORCE NATIVE PATH DELIMITER
		shownName = $.@(fn)
	 end if
	me.pMainWindow.setWindowTitle(me.pAppName&" - "&shownName&"[*]")
	me.pMainWindow.setWindowModified(FALSE)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileNew (me)
	if me.maybeSave() then
		me.pEditor.clear()
		me.setCurrentFileName()
	end if
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileOpen (me)
	filter = "All Files (*)"

	-- ADD FILTER FOR ALL FILETYPES SUPPORTED BY LEXERS
	str = ""
	cnt = me.pLangDict.count
	repeat with i = 1 to cnt
		put "*."&me.pLangDict.getPropAt(i)&" " after str
	end repeat
	delete the last char of str
	put ";;Supported ("&str&")" after filter

	fn = $.Qt.FileDialog.getOpenFileName (me.pMainWindow, "Open File...", $.path, filter)
	if fn<>"" then me.load(fn)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileSave (me)
	if voidP(me.pFileName) then return me.fileSaveAs()
	-- NOTICE: ALWAYS SAVES AS UTF-8
	success = $.file.putString(me.pFileName, me.pEditor.text())
	if success then me.pEditor.setModified(FALSE)
	return success
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on fileSaveAs (me)
	filter = "All Files (*)"
	fn = $.Qt.FileDialog.getSaveFileName (me.pMainWindow, "Save as...", $.path&me.pFileName, filter)
	if fn="" then return FALSE
	me.setCurrentFileName(fn)
	return me.fileSave()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on filePrint (me)
	dialog = $.Qt.PrintDialog.new(me.pMainWindow, me.pPrinter)
	dialog.setWindowTitle("Print Document")
	res = dialog.exec()
	if res = $.Qt.QDialog.Accepted then
		me.pEditor.print(me.pPrinter)
	end if
	dialog.destroy()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on filePageSetup (me)
	dialog = $.Qt.PageSetupDialog.new(me.pMainWindow, me.pPrinter)
	dialog.exec()
	dialog.destroy()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on filePrintPreview (me)
	$.Qt.PrintPreviewDialog.show(me.pMainWindow, me.pEditor, me.pPrinter)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on filePrintPdf (me)
	filter = "PDF files (*.pdf);;All Files (*)"
	fn = $.Qt.FileDialog.getSaveFileName (me.pMainWindow, "Export PDF", void, filter)
	if fn="" then return FALSE
	me.pPrinter.setOutputFormat($.Qt.QPrinter.PdfFormat)
	me.pPrinter.setOutputFileName(fn)
	me.pEditor.print(me.pPrinter)
end

--------------------------------------
-- @callback both action and event callback
--------------------------------------
on fileQuit (me)
	if (me.maybeSave()) then
		-- SAVE WINDOW SETTINGS
		settings = $.Qt.Settings.new("Fluxus", me.pAppName, $.Qt.QSettings.IniFormat)
		settings.setValue("mainwindow/rect", string(me.pMainWindow.geometry()))
		settings.setValue("mainwindow/state", string(me.pMainWindow.windowState()))
		me.pEditor.destroy() -- otherwise crash!
		halt()
	end if
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on copyAvailable (me, widgetID, bool)
	me.pActionCut.setEnabled(bool)
	me.pActionCopy.setEnabled(bool)
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on cursorPositionChanged (me, widget, linenum, index)
	--put "cursorPositionChanged", linenum, index
end

--------------------------------------
-- @callback widget callback
--------------------------------------
on modificationChanged (me, widgetID, bool)
	me.pActionSave.setEnabled(bool)
	me.pMainWindow.setWindowModified(bool)
end

--------------------------------------
-- @callback object callback (called by clipboard object)
--------------------------------------
on clipboardDataChanged (me)
	str = me.pClipboard.text()
	me.pActionPaste.setEnabled(str<>"")
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on helpEdit (me)
	if (not me.maybeSave()) then return
	-- SCRIPTDICT ACTUALLY SUPPOSED TO BE PRIVATE, WHATEVER
	fn = $._.scriptdict.getOne(me.script)
	-- SAVE A BACKUP AS "mainwin.ls~"
	$.file.putString(fn&"~", $.file.getString(fn))
	me.load(fn)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on helpAbout (me)
	msg = "QScintilla-based Code-Editor written in Lingo.<br><br>"&\
	"This application is based on a small runtime engine I called 'LSW', "&\
	"it was written in a code editor (UltraEdit) without ever touching the Director IDE.<br><br>"&\
	"The idea of LSW is to write batch scripts (e.g.for daily automation tasks etc.) or "&\
	"desktop applications in Lingo, in a way similar to e.g. Node.js or Python.<br><br>"&\
	"More infos on LSW <a href='http://valentin.dasdeck.com/lingo/lsw/'>here</a>."
	$.Qt.MessageBox.about(me.pMainWindow, "About "&me.pAppName, msg)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on helpAboutQt (me)
	app = $.Qt.Application.new()
	app.aboutQt()
end

----------------------------------------
-- @callback dropHandler
----------------------------------------
on dropHandler (me, args)
	dropData = args[1]
	if ilk(dropData)<>#list then return
	if dropData.count=0 then return
	if (not me.maybeSave()) then return
	fn = dropData[1]
	me.load(fn)
end
