--!parent
--!encoding=utf-8

--****************************************************************************
-- @package   Qt OOP API
-- @file      Widget Interface
-- @version   0.2
-- @author    Valentin Schmidt
--****************************************************************************

global $

property ancestor
property _children
property _listeners
property _eventlisteners

----------------------------------------
-- @constructor
----------------------------------------
on new (me, parentID)

  me._children = []
  me._listeners = [:]
  me._eventlisteners = []

  return me
end

----------------------------------------
-- ACTIONS
----------------------------------------

-- @return list actionIDs
on actions (me)
  ids = $.Qt._qx.callWidget(me._id, #actionIDs)
  res = []
  repeat with id in ids
    cn = $.Qt._qx.callObject(id, #className)
    delete char 1 of cn
    class = value("$.Qt."&cn)
    if ilk(class)=#script then
      obj = class._make(id)
      --me._children.add(obj) -- ???
      res.add(obj)
    end if
  end repeat
  return res
end

on activateWindow (me)
  $.Qt._qx.callWidget(me._id, #activateWindow)
end

on adjustSize (me)
  $.Qt._qx.callWidget(me._id, #adjustSize)
end

-- @param integer x
-- @param integer y
-- @return integer widgetID
on childAt (me, x, y)
  id = $.Qt._qx.callWidget(me._id, #childAt, x, y)
  if id then
    cn = $.Qt._qx.callWidget(id, #className)
    delete char 1 of cn
    class = value("$.Qt."&cn)
    if ilk(class)=#script then
      obj = class._make(id)
      me._children.add(obj)
      return obj
    end if
  end if
end

-- @return list widgetIDs
on children (me,getObjects)
	res = []
	if getObjects=TRUE then
		ids = $.Qt._qx.callWidget(me._id, #children,TRUE)
	  repeat with id in ids
	    cn = $.Qt._qx.callObject(id, #objectClass)
	    class = value("$.Qt."&cn)
	    if ilk(class)=#script then
	      obj = class._make(id)
	      me._children.add(obj) -- ???
	      res.add(obj)
	    end if
	  end repeat
  else
  	ids = $.Qt._qx.callWidget(me._id, #children)
	  repeat with id in ids
	    cn = $.Qt._qx.callWidget(id, #className)
	    delete char 1 of cn
	    class = value("$.Qt."&cn)
	    if ilk(class)=#script then
	      obj = class._make(id)
	      --me._children.add(obj) -- ???
	      res.add(obj)
	    end if
	  end repeat
  end if
  return res
end

-- @return string name
on className (me)
  return $.Qt._qx.callWidget(me._id, #className)
end

on clearFocus (me)
  $.Qt._qx.callWidget(me._id, #clearFocus)
end

on clearMask (me)
  $.Qt._qx.callWidget(me._id, #clearMask)
end

on close (me)
  $.Qt._qx.callWidget(me._id, #close)
end

-- @return rect margins
on contentsMargins (me)
  return $.Qt._qx.callWidget(me._id, #contentsMargins)
end

-- @return rect r
on contentsRect (me)
  return $.Qt._qx.callWidget(me._id, #contentsRect)
end

on deleteLater (me)
  $.Qt._qx.callWidget(me._id, #deleteLater)
end

on destroy (me)
  $.Qt._qx.callWidget(me._id, #destroy)
end

---- @param integer WID
---- @return integer widgetID
--on find (me, WID)
--  id = $.Qt._qx.callWidget(me._id, #find, WID)
--  if id then
--    cn = $.Qt._qx.callWidget(id, #className)
--    delete char 1 of cn
--    class = value("$.Qt."&cn)
--    if ilk(class)=#script then
--      obj = class._make(id)
--      me._children.add(obj)
--      return obj
--    end if
--  end if
--end

-- @param string widgetOrObjectName
-- @param symbol widgetOrObjectClass - optional
-- @param bool isObject - optional

-- @return integer widgetOrObjectID
on findChild (me, widgetOrObjectName, widgetOrObjectClass, isObject)
	if the paramCount>3 then
  	id = $.Qt._qx.callWidget(me._id, #findChild, widgetOrObjectName, widgetOrObjectClass, isObject)
	else if the paramCount>2 then
  	id = $.Qt._qx.callWidget(me._id, #findChild, widgetOrObjectName, widgetOrObjectClass)
	else
  	id = $.Qt._qx.callWidget(me._id, #findChild, widgetOrObjectName)
  end if
  if id then
    cn = $.Qt._qx.callWidget(id, #className)
    if cn starts "My" then delete char 1 to 2 of cn
    else delete char 1 of cn
    class = value("$.Qt."&cn)
    if ilk(class)=#script then
      obj = class._make(id)
      me._children.add(obj)
      return obj
    end if
  end if
end

-- @param string widgetName - optional
-- @param symbol widgetClass - optional
-- @param bool isObject - optional

-- @return list widgetOrObjectIDs
on findChildren (me, widgetName, widgetClass, isObject)
	if the paramCount>3 then
  	ids = $.Qt._qx.callWidget(me._id, #findChildren, widgetName, widgetClass, isObject)
	else if the paramCount>2 then
  	ids = $.Qt._qx.callWidget(me._id, #findChildren, widgetName, widgetClass)
	else if the paramCount>1 then
  	ids = $.Qt._qx.callWidget(me._id, #findChildren, widgetName)
	else
  	ids = $.Qt._qx.callWidget(me._id, #findChildren)
  end if
    
  res = []
  
  if isObject then
  	 repeat with id in ids
	    cn = $.Qt._qx.callObject(id, #objectClass)
			class = value("$.Qt."&cn)
	    if ilk(class)=#script then
	      obj = class._make(id)
	      --me._children.add(obj) -- ???
	      res.add(obj)
	    end if
	  end repeat
	  
  else
	  repeat with id in ids
	    cn = $.Qt._qx.callWidget(id, #className)
	    delete char 1 of cn
	    class = value("$.Qt."&cn)
	    if ilk(class)=#script then
	      obj = class._make(id)
	      --me._children.add(obj) -- ???
	      res.add(obj)
	    end if
	  end repeat
  end if
  

  
  return res
end

-- @return integer widgetID
on focusWidget (me)
  --return $.Qt._qx.callWidget(me._id, #focusWidget)
  id = $.Qt._qx.callWidget(me._id,#focusWidget)
  if id then
    cn = $.Qt._qx.callWidget(id, #className)
    delete char 1 of cn
    class = value("$.Qt."&cn)
    if ilk(class)=#script then
      obj = class._make(id)
      me._children.add(obj)
      return obj
    end if
  end if
end

-- @return integer objectID
on getFont (me)
  --return $.Qt._qx.callWidget(me._id, #getFont)
  obj = $.Qt.Font._make($.Qt._qx.callWidget(me._id,#getFont))
  me._children.add(obj)
  return obj
end

-- @return integer fontMetricsID
on getFontMetrics (me)
  --return $.Qt._qx.callWidget(me._id, #getFontMetrics)
  obj = $.Qt.FontMetrics._make($.Qt._qx.callWidget(me._id,#getFontMetrics))
  me._children.add(obj)
  return obj
end

-- @return rect r
on frameGeometry (me)
  return $.Qt._qx.callWidget(me._id, #frameGeometry)
end

-- @return rect r
on geometry (me)
  return $.Qt._qx.callWidget(me._id, #geometry)
end

-- @return integer objectID
on getPalette (me)
	obj = $.Qt.Palette._make($.Qt._qx.callWidget(me._id,#getPalette))
  --me._children.add(obj)
  return obj
  --return $.Qt._qx.callWidget(me._id, #getPalette)
end

-- @return integer gestureType
on grabGesture (me, gestureType)
  return $.Qt._qx.callWidget(me._id, #grabGesture)
end

on grabKeyboard (me)
  $.Qt._qx.callWidget(me._id, #grabKeyboard)
end

on grabMouse (me)
  $.Qt._qx.callWidget(me._id, #grabMouse)
end

-- @param string keySequence
on grabShortcut (me, keySequence)
  return $.Qt._qx.callWidget(me._id, #grabShortcut, keySequence)
end

-- @return integer h
on height (me)
  return $.Qt._qx.callWidget(me._id, #height)
end

on hide (me)
  $.Qt._qx.callWidget(me._id, #hide)
end

-- @return bool flag
on isFullScreen (me)
  return $.Qt._qx.callWidget(me._id, #isFullScreen)
end

-- @return bool flag
on isHidden (me)
  return $.Qt._qx.callWidget(me._id, #isHidden)
end

-- @return bool flag
on isMaximized (me)
  return $.Qt._qx.callWidget(me._id, #isMaximized)
end

-- @return bool flag
on isMinimized (me)
  return $.Qt._qx.callWidget(me._id, #isMinimized)
end

-- @return bool flag
on isVisible (me)
  return $.Qt._qx.callWidget(me._id, #isVisible)
end

-- @return integer objectID
on layout (me)
  return $.Qt._qx.callWidget(me._id, #layout)
end

on lower (me)
  $.Qt._qx.callWidget(me._id, #lower)
end

-- @return point hint
on minimumSizeHint (me)
  return $.Qt._qx.callWidget(me._id, #minimumSizeHint)
end

-- @param integer x
-- @param integer y
on move (me, x, y)
  $.Qt._qx.callWidget(me._id, #move, x, y)
end

-- @return string name
on objectName (me)
  return $.Qt._qx.callWidget(me._id, #objectName)
end

-- @return integer widgetID
on parentWidget (me)
  id = $.Qt._qx.callWidget(me._id, #parentWidget)
  if id then
    cn = $.Qt._qx.callWidget(id, #className)
    delete char 1 of cn
    class = value("$.Qt."&cn)
    if ilk(class)=#script then
      obj = class._make(id)
      return obj
    end if
  end if
end

-- @return point p
on pos (me)
  return $.Qt._qx.callWidget(me._id, #pos)
end

on raise (me)
  $.Qt._qx.callWidget(me._id, #raise)
end

-- @return rect r
on rect (me)
  return $.Qt._qx.callWidget(me._id, #rect)
end

on releaseKeyboard (me)
  $.Qt._qx.callWidget(me._id, #releaseKeyboard)
end

on releaseMouse (me)
  $.Qt._qx.callWidget(me._id, #releaseMouse)
end

-- removeAction (int objectID)
on removeAction (me,p1)
  if ilk(p1)=#instance then p1 = p1._id
  me._children.deleteOne(p1) -- ???
  $.Qt._qx.callWidget(me._id,#removeAction,p1)
end

-- @param integer w
-- @param integer h
on resize (me, w, h)
  $.Qt._qx.callWidget(me._id, #resize, w, h)
end

-- @param integer dx
-- @param integer dy
on scroll (me, dx, dy)
  $.Qt._qx.callWidget(me._id, #scroll, dx, dy)
end

-- @param bool flag
on setAcceptDrops (me, flag)
  $.Qt._qx.callWidget(me._id, #setAcceptDrops, flag)
end

-- @param integer attr
-- @param bool flag
on setAttribute (me, attr, flag)
  $.Qt._qx.callWidget(me._id, #setAttribute, attr, flag)
end

-- @param integer left
-- @param integer top
-- @param integer right
-- @param integer bottom
on setContentsMargins (me, left, top, right, bottom)
  $.Qt._qx.callWidget(me._id, #setContentsMargins, left, top, right, bottom)
end

-- @param integer policy
on setContextMenuPolicy (me, policy)
  $.Qt._qx.callWidget(me._id, #setContextMenuPolicy, policy)
end

-- @param integer cur
on setCursor (me, cur)
  $.Qt._qx.callWidget(me._id, #setCursor, cur)
end

-- @param bool flag
on setDisabled (me, flag)
  $.Qt._qx.callWidget(me._id, #setDisabled, flag)
end

-- @param integer h
on setFixedHeight (me, h)
  $.Qt._qx.callWidget(me._id, #setFixedHeight, h)
end

-- @param integer w
-- @param integer h
on setFixedSize (me, w, h)
  $.Qt._qx.callWidget(me._id, #setFixedSize, w, h)
end

-- @param integer w
on setFixedWidth (me, w)
  $.Qt._qx.callWidget(me._id, #setFixedWidth, w)
end

-- @param integer reason - optional
on setFocus (me, reason)
  $.Qt._qx.callWidget(me._id, #setFocus, reason)
end

-- @param integer policy
on setFocusPolicy (me, policy)
  $.Qt._qx.callWidget(me._id, #setFocusPolicy, policy)
end

---setFont (int objectID)
---setFont (string fontDescription)
---setFont (string family, float|int size[, int weight, bool italic])

---- @param integer FontID
on setFont (me, p1,p2,p3,p4)
  pc = the paramCount
  case (pc) of
    2: 	
    	if ilk(p1)=#instance then p1=p1._id
    	$.Qt._qx.callWidget(me._id, #setFont, p1)
    3: $.Qt._qx.callObject(me._id,#setFont, p1,p2)
    4: $.Qt._qx.callObject(me._id,#setFont, p1,p2,p3)
    5: $.Qt._qx.callObject(me._id,#setFont, p1,p2,p3,p4)
  end case
end

-- @param rect r
on setGeometry (me, r)
  $.Qt._qx.callWidget(me._id, #setGeometry, r)
end

-- @param symbol effectType
on setGraphicsEffect (me, effectType) -- , ...
  $.Qt._qx.callWidget(me._id, #setGraphicsEffect, effectType)
end

-- @param integer objectID
on setLayout (me, layout)
  $.Qt._qx.callWidget(me._id, #setLayout, layout._id)
end

-- @param rect|list|image mask
on setMask (me, mask)
  $.Qt._qx.callWidget(me._id, #setMask, mask)
end

-- @param bool on
on setMouseTracking (me,p1)
  $.Qt._qx.callWidget(me._id,#setMouseTracking,p1)
end

-- @param integer w
-- @param integer h
on setMaximumSize (me, w, h)
  $.Qt._qx.callWidget(me._id, #setMaximumSize, w, h)
end

-- @param integer w
-- @param integer h
on setMinimumSize (me, w, h)
  $.Qt._qx.callWidget(me._id, #setMinimumSize, w, h)
end

-- @param string name
on setObjectName (me, name)
  $.Qt._qx.callWidget(me._id, #setObjectName, name)
end

-- @param integer objectID
on setPalette (me, p1)
  if ilk(p1)=#instance then
    --p1._children.add(me)
    p1 = p1._id
  end if
  $.Qt._qx.callWidget(me._id, #setPalette,p1)
end

-- @param integer widgetID
on setParent (me,p1)
  if ilk(p1)=#instance then
    --p1._children.add(me)
    p1 = p1._id
  end if
  $.Qt._qx.callWidget(me._id, #setParent,p1)
end

-- @param integer horizontal
-- @param integer vertical
on setSizePolicy (me, horizontal, vertical)
  $.Qt._qx.callWidget(me._id, #setSizePolicy, horizontal, vertical)
end

-- @param string tip
on setStatusTip (me, tip)
  $.Qt._qx.callWidget(me._id, #setStatusTip, tip)
end

-- @param symbol style
-- supported styles: #CleanlooksStyle, #MotifStyle, #PlastiqueStyle, #WindowsStyle, #WindowsVistaStyle, #WindowsXPStyle
on setStyle (me, style)
  $.Qt._qx.callWidget(me._id, #setStyle, style)
end

-- @param string styleSheet
on setStyleSheet (me, styleSheet)
  $.Qt._qx.callWidget(me._id, #setStyleSheet, styleSheet)
end

-- @param integer firstWidgetID
-- @param integer secondWidgetID
on setTabOrder (me, firstWidgetID, secondWidgetID)
  $.Qt._qx.callWidget(me._id, #setTabOrder, firstWidgetID, secondWidgetID)
end

-- @param string tip
on setToolTip (me, tip)
  $.Qt._qx.callWidget(me._id, #setToolTip, tip)
end

-- @param bool flag
on setUpdatesEnabled (me, flag)
  $.Qt._qx.callWidget(me._id, #setUpdatesEnabled, flag)
end

-- @param bool flag
on setVisible (me, flag)
  $.Qt._qx.callWidget(me._id, #setVisible, flag)
end

-- @param string txt
on setWhatsThis (me, txt)
  $.Qt._qx.callWidget(me._id, #setWhatsThis, txt)
end

-- @param integer flags
on setWindowFlags (me, flags)
  $.Qt._qx.callWidget(me._id, #setWindowFlags, flags)
end

-- @param string|image icon
on setWindowIcon (me, icon)
  $.Qt._qx.callWidget(me._id, #setWindowIcon, icon)
end

-- @param integer modality
on setWindowModality (me, modality)
  $.Qt._qx.callWidget(me._id, #setWindowModality, modality)
end

-- @param bool flag
on setWindowModified (me, flag)
  $.Qt._qx.callWidget(me._id, #setWindowModified, flag)
end

-- @param integer state
on setWindowState (me, state)
  $.Qt._qx.callWidget(me._id, #setWindowState, state)
end

-- @param string title
on setWindowTitle (me, title)
  $.Qt._qx.callWidget(me._id, #setWindowTitle, title)
end

on show (me)
  $.Qt._qx.callWidget(me._id, #show)
end

on showFullScreen (me)
  $.Qt._qx.callWidget(me._id, #showFullScreen)
end

on showMaximized (me)
  $.Qt._qx.callWidget(me._id, #showMaximized)
end

on showMinimized (me)
  $.Qt._qx.callWidget(me._id, #showMinimized)
end

on showNormal(me)
  $.Qt._qx.callWidget(me._id, #showNormal)
end

-- @return point p
on size (me)
  return $.Qt._qx.callWidget(me._id, #size)
end

-- @return point p
on sizeHint (me)
  return $.Qt._qx.callWidget(me._id, #sizeHint)
end

-- @return string tip
on statusTip (me)
  return $.Qt._qx.callWidget(me._id, #statusTip)
end

-- @return bool flag
on testAttribute (me, attr)
  return $.Qt._qx.callWidget(me._id, #testAttribute, attr)
end

-- @return bool flag
on underMouse (me)
  return $.Qt._qx.callWidget(me._id, #underMouse)
end

-- @return symbol class
on widgetClass (me)
  return $.Qt._qx.callWidget(me._id, #widgetClass)
end

-- @return integer w
on width (me)
  return $.Qt._qx.callWidget(me._id, #width)
end

-- @return integer flags
on windowFlags (me)
  return $.Qt._qx.callWidget(me._id, #windowFlags)
end

-- @return image icon
on windowIcon (me)
  return $.Qt._qx.callWidget(me._id, #windowIcon)
end

-- @return integer modality
on windowModality (me)
  return $.Qt._qx.callWidget(me._id, #windowModality)
end

-- @return integer state
on windowState (me)
  return $.Qt._qx.callWidget(me._id, #windowState)
end

-- @return string title
on windowTitle (me)
  return $.Qt._qx.callWidget(me._id, #windowTitle)
end

-- @return integer winId
on winId (me)
  return $.Qt._qx.callWidget(me._id, #winId)
end

-- @return integer x
on x (me)
  return $.Qt._qx.callWidget(me._id, #x)
end

-- @return integer y
on y (me)
  return $.Qt._qx.callWidget(me._id, #y)
end

----------------------------------------
-- STATIC
----------------------------------------

-- tr (string sourceText, *string disambiguation, int n) -> string
--on tr (scr,p1,p2,p3)
--	if the paramCount>3 then
--  	return $.Qt._qx.callWidget(me._class,#tr,p1,p2,p3)
--	else if the paramCount>2 then
--  	return $.Qt._qx.callWidget(me._class,#tr,p1,p2)
--	else
--  	return $.Qt._qx.callWidget(me._class,#tr,p1)
--  end if
--end

----------------------------------------
-- SIGNALS
----------------------------------------

-- #customContextMenuRequested -> point
on customContextMenuRequested (me,id,p1)
  me._detach(#customContextMenuRequested,me,p1)
end

----------------------------------------
-- me.connect(signal)
-- me.connect(signal, receiver)
-- me.connect(signal, receiver, method)
--
-- @return {bool}
----------------------------------------
on connect (me, signal, receiver, method, blockEvent)
  if voidP(receiver) then receiver = _movie
  
	ok = FALSE
	
	-- EVENT MODE
  if integerP(signal) then
    obj = $.include($.Qt.libDir&"classes/signal.ls").new(me, method, receiver,   signal)
    if voidP(blockEvent) then blockEvent = FALSE
    $.Qt._qx.setEventCallback(me._id, signal, #_detach, obj, blockEvent)
    me._eventlisteners.add(obj)
    return TRUE
  end if

  if voidP(method) then method = signal
  if voidP(me._listeners[signal]) then me._listeners[signal] = []

  c = [receiver, method] -- connection target

  -- dont allow SAME connection twice?
  if me._listeners[signal].getPos(c) then return FALSE

  me._listeners[signal].add(c)

  if count(me._listeners[signal])=1 then
    ok = $.Qt._qx.setWidgetCallback(me._id, signal, signal, me) -- connect
  end if

  return ok
end

----------------------------------------
-- me.disconnect() -> Disconnect everything connected to object
-- me.disconnect(signal) -> Disconnect everything connected to object's <signal>
-- me.disconnect(signal, receiver) -> Disconnect <receiver> connected to object's <signal>
-- me.disconnect(signal, receiver, method)  -> Disconnect [<receiver>, <method>] connected to object's <signal>

-- @return {bool}
----------------------------------------
on disconnect (me, signal, receiver, method)

	-- EVENT MODE
  if integerP(signal) then
  	$.Qt._qx.setEventCallback(me._id, signal)
  	-- remove corresponding objects from _eventlisteners list
  	cnt = me._eventlisteners.count
  	repeat with i = cnt down to 1
  		obj = me._eventlisteners[i]
  		if obj._signal=signal then me._eventlisteners.deleteAt(i)
  	end repeat
  	return TRUE
  end if
  
  -- Disconnect everything connected to object
  if voidP(signal) then
    cnt = count(me._listeners)
    repeat with i = 1 to cnt
      signal = me._listeners.getPropAt(i)
      --$.Qt._qx.setWidgetCallback(me._id, signal) -- disconnect (TODO)
    end repeat
    me._listeners = [:]
    return TRUE -- only return true if anything was deleted?
  end if

  -- nothing connected to <signal>, so exit
  if voidP(me._listeners[signal]) then return FALSE
  if count(me._listeners[signal])=0 then return FALSE

  -- Disconnect everything connected to object's <signal>
  if voidP(receiver) then
    --$.Qt._qx.setWidgetCallback(me._id, signal) -- disconnect (TODO)
    me._listeners[signal] = []
    return TRUE -- only return true if anything was deleted?
  end if

  -- Disconnect <receiver> connected to object's <signal>
  if voidP(method) then
    cnt = count(me._listeners[signal])
    repeat with i = cnt down to 1
      if me._listeners[signal][1]=receiver then me._listeners[signal].deleteAt(i)
    end repeat
    if count(me._listeners[signal])=0 then
      --$.Qt._qx.setWidgetCallback(me._id, signal) -- disconnect (TODO)
    end if
    return TRUE -- only return true if anything was deleted?
  end if

  -- Disconnect [<receiver>: <method>] connected to object's <signal>
  c = [receiver, method] -- connection target
  res = me._listeners[signal].deleteOne(c)
  if count(me._listeners[signal])=0 then
    --$.Qt._qx.setWidgetCallback(me._id, signal) -- disconnect (TODO)
  end if
  return res
end

----------------------------------------
-- Using 1 ms single-shot timeouts to create new threads,
-- so scripting errors in callbacks don't fail silently.
-- @callback
----------------------------------------
on _detach (me, signal, id,v1,v2)
  $.detacher.detach(#_call, me, [signal,me,v1,v2])
end

----------------------------------------
-- We are detached from original thread, now handle signal
-- @callback
----------------------------------------
on _call (me, args)
  listeners = me._listeners[args[1]]
  if not voidP(listeners) then
    repeat with l in listeners
      call(l[2],l[1], args[2],args[3],args[4])
    end repeat
  end if
end
