--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      LSW CORE LIB: STRING
-- @author    Valentin Schmidt
-- @version   0.5
-- @desc      Utility functions for string handling
--****************************************************************************

----------------------------------------
-- Replace in string
-- @param {string} stringToFind
-- @param {string} stringToInsert
-- @param {string} input
-- @return {string}
----------------------------------------
on replace (me, stringToFind, stringToInsert, input)
  output = ""
  findLen = stringToFind.length - 1
  repeat while true
    currOffset = offset(stringToFind, input)
    if currOffset=0 then exit repeat
    put input.char [1..currOffset] after output
    delete the last char of output
    put stringToInsert after output
    delete input.char [1.. (currOffset + findLen)]
  end repeat
  put input after output
  return output
end

----------------------------------------
-- Lower to upper case (only for characters in Latin-1/ISO 8859-1 range)
-- @param {string} str
-- @return {string}
----------------------------------------
on toUpper (me, input)
  alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÜ"
  n=input.length
  out = ""
  repeat with i = 1 to n
    c=input.char[i]
    pos = offset(c,alphabet)
    if pos > 0 then put alphabet.char[pos] after out
    else if c="ß" then put "SS" after out
    else put c after out
  end repeat
  return out
end

----------------------------------------
-- Upper to lower case (only for characters in Latin-1/ISO 8859-1 range)
-- @param {string} input
-- @return {string}
----------------------------------------
on toLower (me, input)
  alphabet = "abcdefghijklmnopqrstuvwxyzäöü"
  n=input.length
  repeat with i = 1 to n
    pos = offset(input.char[i],alphabet)
    if pos > 0 then put alphabet.char[pos] into char i of input
  end repeat
  return input
end

----------------------------------------
-- Remove trailing white space
-- @param {string} aString
-- @param {string} [white]
-- @return {string}
----------------------------------------
on rtrim (me, aString, white)
  if voidP(white) then
    white = [SPACE,TAB,NumToChar(10),NumToChar(13)]
  else if stringP(white) then
    white = [white]
  end if
  repeat while white.getPos(the last char of aString)>0
    delete the last char of aString
  end repeat
  return aString
end

----------------------------------------
-- Remove leading white space
-- @param {string} aString
-- @param {string} [white]
-- @return {string}
----------------------------------------
on ltrim (me, aString, white)
  if voidP(white) then
    white = [SPACE,TAB,NumToChar(10),NumToChar(13)]
  else if stringP(white) then
    white = [white]
  end if
  white = [SPACE,TAB,NumToChar(10),NumToChar(13)]
  repeat while white.getPos(aString.char[1])>0
    delete char 1 of aString
  end repeat
  return aString
end

----------------------------------------
-- Remove leading and trailing white space
-- @param {string} aString
-- @param {string} [white]
-- @return {string}
----------------------------------------
on trim (me, aString, white)
  if voidP(white) then
    white = [SPACE,TAB,NumToChar(10),NumToChar(13)]
  else if stringP(white) then
    white = [white]
  end if
  repeat while white.getPos(the last char of aString)>0
    delete the last char of aString
  end repeat
  repeat while white.getPos(aString.char[1])>0
    delete char 1 of aString
  end repeat
  return aString
end

----------------------------------------
-- Explode string into list
-- @param {string} delim
-- @param {string} str
-- @param {integer} [limit]
-- @return {list}
----------------------------------------
on explode (me, delim, str, limit)
  if delim.length=1 then return me.explodeBySingleChar(delim, str, limit)
  l=[]
  if voidP(str) then return l
  if voidP(limit) then limit=the maxinteger
  dl=length(delim)
  repeat while TRUE
    pos=offset(delim,str)
    if pos=0 then exit repeat
    if pos>1 then
	    s=str.char[1..pos-1]
	    l.add(s)
	  else
	  	l.add("")
	  end if
    delete char 1 to pos+dl-1 of str
    if l.count=limit-1 then
      l.add(str)
      return l
    end if
  end repeat
  if pos=0 then pos=1-dl
  l.add(str.char[pos+dl..str.length])
  return l
end

----------------------------------------
-- Explode string into list based on single char delimiter
-- @param {string} delim
-- @param {string} str
-- @param {integer} [limit]
-- @return {list}
----------------------------------------
on explodeBySingleChar (me, delim, str, limit)
  l = []
  if voidP(str) then return l
  if voidP(limit) then limit=the maxinteger
  od = _player.itemDelimiter
  _player.itemDelimiter = delim
  cnt = str.item.count
  repeat with i = 1 to cnt
    if i=limit then
      l.add(str.item[i..cnt])
      exit repeat
    end if
    l.add(str.item[i])
  end repeat
  _player.itemDelimiter = od
  return l
end

----------------------------------------
-- Explode string into list based on SPACE
-- @param {string} str
-- @return {list}
----------------------------------------
on words (me, str)
	l = []
	if not stringP(str) then return l
	cnt = str.word.count
	repeat with i = 1 to cnt
		l.add(str.word[i])
	end repeat
	return l
end
		
----------------------------------------
-- Copied from $.list (for convenience)
-- @param {string} delim
-- @param {list} l
-- @return {string}
----------------------------------------
on implode (me, delim, l)
  str=""
  repeat with i=1 to l.count
    put l[i]&delim after str
  end repeat
  delete char (str.length-delim.length+1) to str.length of str
  return str
end

----------------------------------------
-- Case-sensitive string comparisson
-- @param {string} str1
-- @param {string} str2
-- @return {bool}
----------------------------------------
on strcmp (me, str1, str2)
  --  cnt = str1.length
  --  if cnt<>str2.length then return 0
  --  repeat with i = 1 to cnt
  --    if chartonum(str1.char[i])<>chartonum(str2.char[i]) then return 0
  --  end repeat
  --  return 1
  return [str1].getPos(str2)
end

----------------------------------------
-- Find position of sub string, case-sensitive
-- int strpos ( string haystack, string needle [, int offset])
-- @param {string} haystack
-- @param {string} needle
-- @param {string} str1
-- @param {integer} [startPos=1]
-- @return {integer}
----------------------------------------
on strpos (me, haystack, needle, startPos)
  if voidP(startPos) then startPos=1
  haystackLen = haystack.length
  needleLen = needle.length
  needleList = []
  repeat with i = 1 to needleLen
    needleList[i] = chartonum(needle.char[i])
  end repeat
  cnt = haystackLen - needleLen + 1
  repeat with i = startPos to cnt
    n=0
    repeat while (chartonum(haystack.char[i+n]) = needleList[1+n])
      n=n+1
      if n=needleLen then return i
    end repeat
  end repeat
  return 0
end

----------------------------------------
-- Pads string with spaces
-- @param {string} input
-- @param {integer} pad_length
-- @return {string}
----------------------------------------
on strpad (me, input, pad_length)
  len = input.length
  if len>=pad_length then return input
  repeat with i = len+1 to pad_length
    put " " after input
  end repeat
  return input
end
